:- module( by_unix, [    
					(@)/1,
					(@@)/2,
					(@@)/3,
					which/2,
					cd/2,
					by_unix_retract/0,
					by_unix_assert/0,
					by_unix_version/2,
					by_unix_term_to_serial/2,
					op( 200, fy, @ ),
					op( 200, yfx, @@ ),
					op( 200, fy, -- ),
					op( 400, fx, / ),
					op( 400, fx, './' ),
   					op( 400, fx, '../' ),
					op( 600, yfx, '/../' )
				] ).

:- ensure_loaded( library(debug) ).

/** <module> by_unix

	An elegance layer to calling unix commands.

This library provides primitives that allow programmers and users
to embed calls to process_create/3. The aim is to keep application
code clear and succinct. This is achieved by (a) reducing the call to
process_create/3 to its essential constituents and (b) allowing for 
term structures in the arguments of the call. 

A simple example is

== 
  ?- @ ls.
==
which lists all files by translating ls to process_create( path(ls), [], [] ).

To list all details (long list): 
==
	?- @ ls(-l).
==

which demonstrates translation of terms in arguments. 

By_unix looks at the arguments of the terms right of an @ to decide which ones are options
(3rd argument of process_create/3) assuming the rest to be arguments to the call (2nd argument of process_create/3).
Command arguments can be terms which will be serialised, so that a/b becomes 'a/b'. 
The argument * is special and it is expanded fed into expand_file_name/2.

With SWI 7 you can now have '.' in atoms, making interactions with the OS even smoother. 

==
?- @ mkdir( -p, /tmp/test_by_unix ).
?- @ cd( /tmp/test_by_unix ).
?- @ touch( empty.pl ).
?- @ rm( -f, empty.pl ).

?- @ cd( pack(by_unix) ).
?- Wc @@ wc( -l, pack.pl ).

Wc = ['10 pack.pl'].

?- @ cd( @ '$HOME' ).
?- [Pwd] @@ pwd.
Pwd = '/home/nicos'.

==

The main objective of by_unix is achieved by what has been described so far. 
We have found that more than 90 percent of the its uses are to produce elegant Goals that 
are clear to read and construct their arguments within Prolog. 
We provide some more features, which are described in what follows, but they should be seen as marginal.

Changing directory is not supported via cd in process_create as =|which cd|= fails to return
an executable in bash. That is, 

== 
?- process_create( path(cd), ['/'], [] ).
ERROR: source_sink `path(cd)' does not exist
==

As oppose to: 

==
?- [library(by_unix)].

?- @ cd( / ).
true.

?- @ ls.
bin   dev  home        initrd.img.old  lib64	   media  opt	root  sbin     srv  tmp  var
boot  etc  initrd.img  lib	       lost+found  mnt	  proc	run   selinux  sys  usr  vmlinuz
true.

?- @ cd( /home/nicos ).
?- @ cd( pack(by_unix) ).

== 

which/2 provides a locator for executables. by_unix_term_to_serial/2 serialises Prolog terms
to process_create/3 atoms, and by_unix_assert/0 allows for doing away with the @.

As process_create/3 quotes special characters, for instance 
==
?- process_create( path(ls), ['*'], [] ).
/bin/ls: cannot access *: No such file or directory
ERROR: Process "/bin/ls": exit status: 2
==

By_unix allows for in-argument file name expansions via expand_file_name/2. 

?- @ ls( -l, @('*.pl') ).

@@/2 provides dual functionality: either picking up output lines from the calling command
or for maplist/2 applications. See @@/2 and @@/3.

 A lines example 
==
?- @ cd( @ '$HOME' ).
?- Pwd @@ pwd.
Pwd = ['/home/nicos'].
==

 A maplist example 
==
?- @ cd( pack(by_unix) ).
?- Files = ['by_unix.pl'], maplist( @@wc(-l), Files, Wcs ).

==

	@author Nicos Angelopoulos
	@version 0.1.6   2013/12/26
	@see http://stoics.org.uk/~nicos/sware/by_unix

*/

%% @@( -Lines, +Comm ).
%% @@( +Comm, -Arg ).
%
%  If first argument is a variable or list, it is interpretted to be the Lines invocation,
%  Output from Comm instantiated in Lines. Attach Arg to Comm before processing as a 
%  Unix command.
@@(Lines,Goal) :-
	once( var(Lines); is_list(Lines) ),
	!,
	by_unix_separate( Goal, Name, TArgs, Opts ),
	\+ memberchk( stdout(_), Opts  ),
	unix_process( Name, Goal, TArgs, [stdout(pipe(Out))|Opts] ),
	read_lines(Out, Lines).
@@(Goal,Arg) :-
	by_unix_separate( Goal, Name, Args, Opts ),
	by_unix_term_to_serial( Arg, Serial ),
	to_list( Serial, Serials ),
	append( Args, Serials, All ),
	% process_create( path(Name), All, Opts ).
	unix_process( Name, Goal, All, Opts ).
% this is suitable for meta calls, with output
%% @@( +Comm, +Arg, -Lines ).
%
% Attach Arg to Comm before processing as a Unix command and provide output to Lines.
% Works with maplist/3 but Lines will be triply nested.
%
@@(Goal,Arg,Lines) :-
	by_unix_separate( Goal, Name, Args, Opts ),
	by_unix_term_to_serial( Arg, Serial ),
	to_list( Serial, Serials ),
	append( Args, Serials, All ),
	process_create( path(Name), All, [stdout(pipe(Out))|Opts] ),
	read_lines(Out, Lines ).

%% @(+Goal).
%
%  This is the main predicate of by_unix.
%  See module documentation for examples. 
%
%  For @cd( Arg ) see documentation of cd/2.
% ==
%    ?- @ mkdir( -p, /tmp/test_by_unix ).
%    ?- @ cd( /tmp/test_by_unix ).
%    ?- @ touch( empty.pl ).
%    ?- @ rm( -f, empty.pl ).
% ==
% 
@(Goal) :-
	by_unix_separate( Goal, Name, TArgs, Opts ),
	unix_process( Name, Goal, TArgs, Opts ).

by_unix_separate( Goal, Name, TArgs, Opts ) :-
	% Goal =.. [Name|GArgs],
	compound( Goal, Name, GArgs ),

	which( Name, _Wch ), %fixme add error
	partition( pc_option, GArgs, Opts, Args ),
	maplist( by_unix_term_to_serial, Args, NesTArgs ),
	flatten( NesTArgs, TArgs ).

unix_process( Cd, Goal, [_Arg], [] ) :-
	Cd == cd,
	!,
	arg( 1, Goal, Garg ),
	cd( Garg ).
unix_process( Name, _Goal, Args, Opts ) :-
	debug( by_unix, 'Sending, name: ~w, args: ~w, opts:~w.', [Name,Args,Opts] ),
	process_create( path(Name), Args, Opts ).

/** which( +Which, -This ).

	Expand Which as a unix command in the path and return its absolute_file_name/3 in This.
	When Which is =|cd|=, variable This is also bound to =|cd|=. cd is handled separately as =|which cd|= 
	fails in bash, as does process_create(path(cd), ['/'], [] ).

*/
which( Which , This ) :- 
	Which == cd,
	!, 
	This = cd.
which( Which, This ) :-
     absolute_file_name( path(Which), This,
			 [ extensions(['',exe]),
			   file_errors(fail),
			   access(exist) % shouldn't this be execute ?
			 ] ).  % does not succeed for built-ins !!!

pc_option( Term ) :-
	compound( Term, Name, Args ),
	length( Args, 1 ),
	% functor( Term, Name, 1 ),
	known_pc_options( OptNames ),
	memberchk( Name, OptNames ).

%% by_unix_term_to_serial( Term, Serial ).
%
%  Term is serialised into an atomic Serial. When Term is *, Serial is the list of current files,
%  while when Term = @ Atom, Serial is the result of applying expand_file_name( Atom, Serial ).
%
by_unix_term_to_serial( *, Files ) :-
	!,
	expand_file_name('*',Files).
by_unix_term_to_serial( @(Atom), Files ) :-
	atomic( Atom ),
	!,
	expand_file_name(Atom,Files).

by_unix_term_to_serial( Term, Serial ) :-
	with_output_to( atom(Serial), write_term(Term,[quoted(false)]) ).

%% by_unix_version( -Version, -Date ).
%
%  Provides version and date of current release.
%
by_unix_version( 0:1:6, date(2013,12,26) ).

%% by_unix_retract.
%
%  Retract all user:goal_expansion(_,_). 
%  
by_unix_retract :-
	Head = user:goal_expansion(_,_),
	retractall( Head ).

%% by_unix_assert.
%
%   Allows for goal expansion of Com to @ Com, when Com is a which-able Unix command.
%   
% ==
%   ?- by_unix_asert. 
%   ?- mkdir( -p, /tmp/test_by_unix ).
%   ?- cd( /tmp/test_by_unix ).
%   ?- touch( empty.pl ).
%   ?- ls( -l ).
%   ?- rm( -f, empty.pl ).
%   ?- ls( -a ).
% ==
by_unix_assert :-
	Head = user:goal_expansion(Term1,Term2),
	Body = (  (atomic(Term1) -> UnixCom = Term1
				; compound_name_arity(Term1,UnixCom,_Arity) ),
			which(UnixCom,_),
			Term2= @(Term1)
		  ),
	assert( (Head :- Body) ).

known_pc_options( [stdin,stdout,stderr,cwd,env,process,detached,window] ).

%% cd( -Old, +New ).
%
%  Similar to working_directory/2, but in addition New can be 
%  a search path alias or the 1st argument of absolute_file_name/2.
%  This is also the case when @ cd( New ) is called.
% 
% == 
% ?- @ cd( pack ).
% true.
% 
% ?- @ pwd.
% /usr/local/users/nicos/local/git/lib/swipl-7.1.4/pack
% true.
% 
% ?- @ cd( @ '$HOME' ).
% true.
% 
% ?- @ pwd.
% /home/nicos
% true.
% == 
% 
cd( Old, New ) :-
	working_directory( Old, Old ),
	cd( New ).

cd( Spec ) :-
	catch(absolute_file_name(Spec,Dir),_,fail),
	exists_directory( Dir ),
	!,
	working_directory( _, Dir ).
cd( Dir ) :-
	ground( Dir ),
	user:file_search_path( Dir, TermLoc ),
	expand_file_search_path( TermLoc, Loc ),
	exists_directory( Loc ),
	!,
	working_directory( _, Loc ).
cd( Atom ) :-
	atom( Atom ),
	!,
	working_directory( _, Atom ).
cd( Term ) :-
	compound( Term ),
	by_unix_term_to_serial( Term, Serial ),
	to_list( Serial, [Dir|_T] ), %fixme: warn if T \== [] 
	working_directory( _, Dir ).

read_lines(Out, Lines) :-
        read_line_to_codes(Out, Line1),
        read_lines(Line1, Out, Lines).
read_lines(end_of_file, _, []) :- !.
read_lines(Codes, Out, [Line|Lines]) :-
        atom_codes(Line, Codes),
        read_line_to_codes(Out, Line2),
        read_lines(Line2, Out, Lines).

compound( Term, Name, Args ) :-
	current_predicate( compound_name_arguments/3 ),
	compound( Term ),  % in real this is after the cut, here we are less strict we allows ls = ls()
	!,
	compound_name_arguments( Term, Name, Args ).
compound( Term, Name, Args ) :-
	Term =.. [Name|Args].

to_list( Serial, Serials ) :-
	is_list( Serial ),
	!,
	Serial = Serials.
to_list( Serial, [Serial] ).
